import collections.abc

import typing_extensions as typing

import sila
from unitelabs.cdk.sila.utils.name import to_identifier

from ..common import Decorator
from ..common.errors import define_error
from ..metadata import Metadatum
from ..utils import parse_docstring, to_display_name

if typing.TYPE_CHECKING:
    from ..common import Feature


class UnobservableCommand(Decorator):
    """
    Any command for which observing the progress of execution is not possible or does not make sense.

    Args:
      name: Human readable name for the command. By default, this is
        automatically inferred by the name of the decorated method.
      identifier: Unique identifier of the command. By default, this
        equals the `name` without spaces and special characters.
      mode: Execution mode of the command. Defines whether subsequent
        executions of this command should be executed in parallel,
        sequentially or not at all.  Defaults to parallel execution mode.
      errors: A list of defined errors that may occur during command
        execution.
      enabled: Callback function that is called to determine whether
        the command is enabled or not. If not provided, the command is
        always enabled.

    Examples:
      Convert a feature method into an unobservable command:
      >>> class MyFeature(sila.Feature):
      ...   @sila.UnobservableCommand()
      ...   async def my_command(self, param_a: str, param_b: int) -> tuple[str, int]:
      ...     \"\"\"
      ...     Describe what your command does.
      ...
      ...     Args:
      ...       ParamA: Describe the purpose of param_a.
      ...       ParamB: Describe the purpose of param_b.
      ...
      ...     Returns:
      ...       ResponseA: Response value a.
      ...       ResponseB: Response value b.
      ...
      ...     Raises:
      ...       RuntimeError: If something goes wrong.
      ...     \"\"\"
      ...     return param_a, param_b
    """

    def __init__(
        self,
        /,
        *,
        name: str | None = None,
        identifier: str | None = None,
        mode: sila.ExecutionMode | None = None,
        errors: collections.abc.Sequence[type[Exception]] | None = None,
        enabled: bool | collections.abc.Callable[..., bool] = True,
    ) -> None:
        super().__init__(identifier=identifier, name=name, errors=errors, enabled=enabled)

        self._mode = mode

    @typing.override
    def attach(self, feature: "Feature") -> bool:
        if not super().attach(feature):
            return False

        docstring = parse_docstring(self._function, feature=feature)

        self._name = self._name or to_display_name(self._function.__name__)
        self._identifier = self._identifier or to_identifier(self._name)
        self._description = docstring.description
        self._parameters = docstring.parameters
        self._responses = docstring.returns
        self._errors = [*self._errors, *docstring.raises.values()]

        self._handler = sila.server.UnobservableCommand(
            identifier=self._identifier,
            display_name=self._name,
            description=self._description,
            execution_mode=self._mode or sila.ExecutionMode.PARALLEL,
            function=self.execute,
            errors={Error.identifier: Error for error in self._errors if (Error := define_error(error))},
            parameters=self._parameters,
            responses=self._responses,
            feature=feature,
        )
        self._metadata = Metadatum._infer_metadata(self)

        return True



import collections.abc
import inspect

import typing_extensions as typing
from sila.server import Native, SiLAError, UndefinedExecutionError

import sila

from ...subscriptions import Subscription
from ..common import Decorator
from ..common.errors import define_error
from ..metadata import Metadatum
from ..utils import parse_docstring, to_display_name, to_identifier

if typing.TYPE_CHECKING:
    from ..common import Feature

T = typing.TypeVar("T")
Stream = collections.abc.AsyncIterator[T]


class ObservableProperty(Decorator):
    """
    A property describes certain aspects of a SiLA server that do not require an action on the SiLA server.

    Args:
      name: Human readable name for the property. By default, this is
        automatically inferred by the name of the decorated method.
      identifier: Unique identifier of the property. By default, this
        equals the `name` without spaces and special characters.
      errors: A list of defined errors that may occur during property
        execution.
      enabled: Callback function that is called to determine whether
        the property is enabled or not. If not provided, the property
        is always enabled.

    Examples:
      Convert a feature method into an observable property:
      >>> class MyFeature(sila.Feature):
      ...   @sila.ObservableProperty()
      ...   async def my_property(self) -> sila.Stream[int]:
      ...     \"\"\"
      ...     Describe what your property does.
      ...
      ...     Raises:
      ...       RuntimeError: If something goes wrong.
      ...     \"\"\"
      ...     yield 0
      ...     yield 1
      ...     yield 2
    """

    @typing.override
    def attach(self, feature: "Feature") -> bool:
        if not super().attach(feature):
            return False

        docstring = parse_docstring(self._function, feature=feature)

        self._name = self._name or to_display_name(self._function.__name__.removeprefix("subscribe_"))
        self._identifier = self._identifier or to_identifier(self._name)
        self._description = docstring.description
        self._errors = [*self._errors, *docstring.raises.values()]

        element = next(iter(docstring.returns.values()))
        element.identifier = self._identifier
        element.display_name = self._name
        element.description = self._description

        self._responses = {self._identifier: element}

        self._handler = sila.server.ObservableProperty(
            identifier=self._identifier,
            display_name=self._name,
            description=self._description,
            function=self.execute,
            errors={Error.identifier: Error for error in self._errors if (Error := define_error(error))},
            data_type=element.data_type,
            feature=feature,
        )
        self._metadata = Metadatum._infer_metadata(self)

        return True

    @typing.override
    async def execute(
        self, metadata: dict[sila.MetadataIdentifier, Native], **parameters
    ) -> collections.abc.AsyncIterator[Native]:
        if not self._feature:
            raise RuntimeError

        try:
            function = self._with_metadata(self._function, metadata)
            function = self._with_parameters(function, parameters)

            responses = self._execute(function)

            async for item in responses:
                yield item
        except SiLAError:
            raise
        except Exception as error:
            import traceback

            traceback.print_exc()

            error_name = error.__name__ if inspect.isclass(error) else error.__class__.__name__
            if any(error_name in e.__name__ for e in self._errors):
                raise define_error(error)(str(error)) from None

            msg = f"{error.__class__.__name__}: {error}"
            raise UndefinedExecutionError(msg) from error

    @typing.override
    async def _execute(self, function: collections.abc.Callable) -> collections.abc.AsyncIterator[Native]:
        responses = function()

        if inspect.iscoroutine(responses):
            responses = await responses

        if isinstance(responses, Subscription):
            async for response in responses:
                yield {self._identifier: response}

            responses.terminate()

        elif inspect.isasyncgen(responses):
            async for response in responses:
                yield {self._identifier: response}

        elif inspect.isgenerator(responses):
            for response in responses:
                yield {self._identifier: response}



import dataclasses
import enum
import sys

import sila

if sys.version_info < (3, 12):
    import typing_extensions as typing
else:
    import typing


T = typing.TypeVar("T", bound=sila.Native)


@dataclasses.dataclass
class Enum(sila.Constrained[str]):
    """Enum data type that converts from and to a python enum."""

    _enum: typing.ClassVar[enum.EnumMeta] = enum.Enum

    @typing.override
    @classmethod
    def create(cls, enum_type: enum.EnumMeta) -> type[typing.Self]:
        data_type = super().create(
            sila.String, [sila.Set(values=[sila.String(cls._name_to_value(name)) for name in enum_type.__members__])]
        )
        data_type._enum = enum_type

        return data_type

    @typing.override
    @classmethod
    async def from_native(
        cls,
        context: "sila.Context",
        value: enum.Enum | None = None,
        /,
        *,
        execution: sila.framework.common.Execution | None = None,
    ) -> typing.Self:
        name = cls._name_to_value(value.name) if value is not None else None

        return await super().from_native(context, name, execution=execution)

    @typing.override
    async def to_native(self, context: "sila.Context", /) -> enum.EnumMeta:
        value = await super().to_native(context)
        name = next(name for name in self._enum.__members__ if self._name_to_value(name) == value)

        return self._enum[name]

    @classmethod
    def _name_to_value(cls, name: str) -> str:
        """Convert the enum name to a native string value."""
        return name.lower().replace("_", " ")



import dataclasses
import textwrap

import pytest

import sila
from sila import Element, Integer, List, Real, String, Structure
from unitelabs.cdk.sila.command.unobservable_command import UnobservableCommand
from unitelabs.cdk.sila.common.feature import Feature
from unitelabs.cdk.sila.data_types.custom_data_type import CustomDataType
from unitelabs.cdk.sila.property.unobservable_property import UnobservableProperty


class TestInherit:
    async def test_should_subclass_custom_data_type(self):
        # Create custom data type
        class TestCustomDataType(CustomDataType):
            pass

        # Assert that the method returns the correct value
        assert TestCustomDataType._identifier == "TestCustomDataType"
        assert TestCustomDataType._name == "Test Custom Data Type"
        assert TestCustomDataType._description == textwrap.dedent("""\
            A custom data type definition that can be reused in multiple places.

            Examples:
              Define a custom data type:
              >>> @dataclasses.dataclass
              ... class MyCustomDataType(sila.CustomDataType):
              ...   \"\"\"Describe what your data type is for.\"\"\"
              ...   param_a: str
              ...   param_b: int
              ...
              ... class MyFeature(sila.Feature):
              ...   @sila.UnobservableProperty()
              ...   async def my_property(self) -> MyCustomDataType:
              ...     \"\"\"Describe what your property does.\"\"\"
              ...     return MyCustomDataType(param_a="Hello, World!", param_b=42)
              ...
              ...   @sila.UnobservableCommand()
              ...   async def my_property(self, my_custom_data: MyCustomDataType) -> None:
              ...     \"\"\"
              ...     Describe what your command does.
              ...
              ...     Args:
              ...       my_custom_data: The custom data type to process.
              ...     \"\"\"
              ...     print(my_custom_data.param_a, my_custom_data.param_b)""")

    async def test_should_dynamically_create_custom_data_type(self):
        # Create custom data type
        TestCustomDataType: type[CustomDataType] = dataclasses.make_dataclass(
            "TestCustomDataType", (), bases=(CustomDataType,)
        )

        # Assert that the method returns the correct value
        assert TestCustomDataType._identifier == "TestCustomDataType"
        assert TestCustomDataType._name == "Test Custom Data Type"

    async def test_should_infer_description_from_docs(self):
        # Create custom data type
        class TestCustomDataType(CustomDataType):
            """A specific custom data type."""

        # Assert that the method returns the correct value
        assert TestCustomDataType._identifier == "TestCustomDataType"
        assert TestCustomDataType._name == "Test Custom Data Type"
        assert TestCustomDataType._description == "A specific custom data type."


class TestAttach:
    async def test_should_raise_on_missing_fields(self):
        # Create custom data type
        feature = Feature()

        class TestCustomDataType(CustomDataType):
            pass

        # Attach
        with pytest.raises(
            ValueError,
            match=(
                r"Could not detect any fields on 'TestCustomDataType'\. "
                r"Did you forget to annotated your data type definition with `@dataclasses.dataclass`?"
            ),
        ):
            TestCustomDataType.attach(feature)

    async def test_should_add_custom_data_type_to_feature(self):
        # Create custom data type
        feature = Feature()

        @dataclasses.dataclass
        class TestCustomDataType(CustomDataType):
            value: str

        # Attach
        TestCustomDataType.attach(feature)

        # Assert that the method returns the correct value
        assert feature.data_type_definitions["TestCustomDataType"] == TestCustomDataType._custom

    async def test_should_infer_basic_property(self):
        # Create custom data type
        feature = Feature()

        @dataclasses.dataclass
        class TestCustomDataType(CustomDataType):
            test_custom_data_type: str

        # Attach
        TestCustomDataType.attach(feature)

        # Assert that the method returns the correct value
        assert TestCustomDataType._custom.data_type == String

    async def test_should_infer_list_property(self):
        # Create custom data type
        feature = Feature()

        @dataclasses.dataclass
        class TestCustomDataType(CustomDataType):
            test_custom_data_type: list[float]

        # Attach
        TestCustomDataType.attach(feature)

        # Assert that the method returns the correct value
        assert TestCustomDataType._custom.data_type() == List.create(Real)()

    async def test_should_infer_structure_property(self):
        # Create custom data type
        feature = Feature()

        @dataclasses.dataclass
        class TestCustomDataType(CustomDataType):
            value: int
            another_value: list[float]

        # Attach
        TestCustomDataType.attach(feature)

        # Assert that the method returns the correct value
        assert (
            TestCustomDataType._custom.data_type()
            == Structure.create(
                {
                    "value": Element(
                        identifier="Value",
                        display_name="Value",
                        data_type=Integer,
                    ),
                    "another_value": Element(
                        identifier="AnotherValue",
                        display_name="Another Value",
                        data_type=List.create(Real),
                    ),
                }
            )()
        )


class TestToNative:
    async def test_should_parse_basic_to_native(self):
        # Create custom data type definition
        @dataclasses.dataclass
        class TestCustomDataType(CustomDataType):
            test_custom_data_type: str

        # Create unobservable property
        def function() -> TestCustomDataType:
            return TestCustomDataType(test_custom_data_type="Hello, World!")

        server = sila.server.Server()
        feature = Feature()
        unobservable_property = UnobservableProperty()
        unobservable_property(function)
        unobservable_property.attach(feature)
        server.register_feature(feature)

        # Execute unobservable property
        assert isinstance(unobservable_property._handler, sila.UnobservableProperty)
        response = await unobservable_property._handler.read()

        # Assert that the method returns the correct value
        assert response == (await TestCustomDataType._custom.from_native(server, "Hello, World!")).encode(number=1)

    async def test_should_parse_list_to_native(self):
        # Create custom data type definition
        @dataclasses.dataclass
        class TestCustomDataType(CustomDataType):
            test_custom_data_type: list[float]

        # Create unobservable property
        def function() -> TestCustomDataType:
            return TestCustomDataType(test_custom_data_type=[1.1, 2.2, 3.3])

        server = sila.server.Server()
        feature = Feature()
        unobservable_property = UnobservableProperty()
        unobservable_property(function)
        unobservable_property.attach(feature)
        server.register_feature(feature)

        # Execute unobservable property
        assert isinstance(unobservable_property._handler, sila.UnobservableProperty)
        response = await unobservable_property._handler.read()

        # Assert that the method returns the correct value
        assert response == (await TestCustomDataType._custom.from_native(server, [1.1, 2.2, 3.3])).encode(number=1)

    async def test_should_parse_structure_to_native(self):
        # Create custom data type definition
        @dataclasses.dataclass
        class TestCustomDataType(CustomDataType):
            value: int
            another_value: list[float]

        # Create unobservable property
        def function() -> TestCustomDataType:
            return TestCustomDataType(value=42, another_value=[1.1, 2.2, 3.3])

        server = sila.server.Server()
        feature = Feature()
        unobservable_property = UnobservableProperty()
        unobservable_property(function)
        unobservable_property.attach(feature)
        server.register_feature(feature)

        # Execute unobservable property
        assert isinstance(unobservable_property._handler, sila.UnobservableProperty)
        response = await unobservable_property._handler.read()

        # Assert that the method returns the correct value
        assert response == (
            await TestCustomDataType._custom.from_native(server, {"value": 42, "another_value": [1.1, 2.2, 3.3]})
        ).encode(number=1)


class TestFromNative:
    async def test_should_parse_native_to_basic(self):
        # Create custom data type definition
        @dataclasses.dataclass
        class TestCustomDataType(CustomDataType):
            test_custom_data_type: str

        # Create unobservable property
        def function(value: TestCustomDataType) -> None:
            assert value == TestCustomDataType(test_custom_data_type="Hello, World!")

        server = sila.server.Server()
        feature = Feature()
        unobservable_property = UnobservableCommand()
        unobservable_property(function)
        unobservable_property.attach(feature)
        server.register_feature(feature)

        # Execute unobservable property
        assert isinstance(unobservable_property._handler, sila.UnobservableCommand)
        await unobservable_property._handler.execute(
            (await TestCustomDataType._custom.from_native(server, "Hello, World!")).encode(number=1)
        )

    async def test_should_parse_list_to_native(self):
        # Create custom data type definition
        @dataclasses.dataclass
        class TestCustomDataType(CustomDataType):
            test_custom_data_type: list[float]

        # Create unobservable property
        def function(value: TestCustomDataType) -> None:
            assert value == TestCustomDataType(test_custom_data_type=[1.1, 2.2, 3.3])

        server = sila.server.Server()
        feature = Feature()
        unobservable_property = UnobservableCommand()
        unobservable_property(function)
        unobservable_property.attach(feature)
        server.register_feature(feature)

        # Execute unobservable property
        assert isinstance(unobservable_property._handler, sila.UnobservableCommand)

        await unobservable_property._handler.execute(
            (await TestCustomDataType._custom.from_native(server, [1.1, 2.2, 3.3])).encode(number=1)
        )

    async def test_should_parse_structure_to_native(self):
        # Create custom data type definition
        @dataclasses.dataclass
        class TestCustomDataType(CustomDataType):
            value: int
            another_value: list[float]

        # Create unobservable property
        def function(value: TestCustomDataType) -> None:
            assert value == TestCustomDataType(value=42, another_value=[1.1, 2.2, 3.3])

        server = sila.server.Server()
        feature = Feature()
        unobservable_property = UnobservableCommand()
        unobservable_property(function)
        unobservable_property.attach(feature)
        server.register_feature(feature)

        # Execute unobservable property
        assert isinstance(unobservable_property._handler, sila.UnobservableCommand)
        await unobservable_property._handler.execute(
            (
                await TestCustomDataType._custom.from_native(server, {"value": 42, "another_value": [1.1, 2.2, 3.3]})
            ).encode(number=1)
        )



import datetime
import unittest.mock
import zoneinfo

import pytest

from unitelabs.cdk import sila
from unitelabs.cdk.sila.constraints import MinimalExclusive


class TestValidate:
    @pytest.mark.parametrize(
        "value,validation,error",
        [
            pytest.param(10, sila.data_types.Integer(11), False, id="valid-int"),
            pytest.param(10, sila.data_types.Integer(10), True, id="invalid-int"),
            pytest.param(10.5, sila.data_types.Real(10.6), False, id="valid-float"),
            pytest.param(10.5, sila.data_types.Real(10.5), True, id="invalid-float"),
            pytest.param(datetime.date(2023, 1, 1), sila.data_types.Date(2023, 1, 2), False, id="valid-date"),
            pytest.param(datetime.date(2023, 1, 1), sila.data_types.Date(2023, 1, 1), True, id="invalid-date"),
            pytest.param(
                datetime.time(12, 0, 0, 0, tzinfo=zoneinfo.ZoneInfo("US/Eastern")),
                sila.data_types.Time(12, 0, 0, 1),
                False,
                id="valid-time",
            ),
            pytest.param(
                datetime.time(12, 0, 0, 0, tzinfo=zoneinfo.ZoneInfo("US/Eastern")),
                sila.data_types.Time(12, 0, 0, 0),
                True,
                id="invalid-time",
            ),
            pytest.param(
                datetime.datetime(2023, 1, 1, 12, 0, 0, 0, tzinfo=zoneinfo.ZoneInfo("US/Eastern")),
                sila.data_types.Timestamp(2023, 1, 1, 12, 0, 0, 1),
                False,
                id="valid-timestamp",
            ),
            pytest.param(
                datetime.datetime(2023, 1, 1, 12, 0, 0, 0, tzinfo=zoneinfo.ZoneInfo("US/Eastern")),
                sila.data_types.Timestamp(2023, 1, 1, 12, 0, 0, 0),
                True,
                id="invalid-timestamp",
            ),
        ],
    )
    async def test_validation(self, value, validation, error):
        constraint = MinimalExclusive(value)
        if not error:
            assert await constraint.validate(validation)
        else:
            with pytest.raises(ValueError):
                await constraint.validate(validation)

    async def test_should_raise_type_error_unsupported_type(self):
        with pytest.raises(TypeError):
            MinimalExclusive(unittest.mock.Mock())



import asyncio
import collections.abc
import time
import weakref

import typing_extensions as typing

from .default import _DEFAULT_VALUE, Default

if typing.TYPE_CHECKING:
    from .subject import Subject

T = typing.TypeVar("T")


class Subscription(asyncio.Queue[T], collections.abc.AsyncIterator[T]):
    """An AsyncIterable you can asynchronously add items to."""

    def __init__(self, maxsize: int, parent: "Subject") -> None:
        super().__init__(maxsize)

        self._parent: Subject = weakref.proxy(parent)
        self._value: T | Default = typing.cast(T, _DEFAULT_VALUE)
        self._closed = asyncio.Event()

    def __repr__(self) -> str:
        name = self.__class__.__name__
        identifier = str(id(self))
        return f"{name}({identifier}, size={self.size}, closed={self._closed.is_set()})"

    @property
    def size(self) -> int:
        """The number of items in the queue."""

        return self.qsize()

    def update(self, value: "T") -> None:
        """Update the current value, if `value` is not current value."""

        if value != self._value:
            self._value = value
            self.put_nowait(value)

    def cancel(self) -> None:
        """Cancel the subscription."""

        self._closed.set()

    def terminate(self) -> None:
        """Unsubscribe the subscription from its parent."""

        self._parent.unsubscribe(self)

    def __aiter__(self) -> collections.abc.AsyncIterator[T]:
        return self

    async def __anext__(self) -> T:
        if self._closed.is_set():
            raise StopAsyncIteration

        cancellation = asyncio.create_task(self._closed.wait(), name="subscription-cancellation")
        queue_task = asyncio.create_task(super().get(), name="subscription-queue")
        tasks: set[asyncio.Task] = {cancellation, queue_task}

        try:
            done, _ = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)

            if cancellation in done:
                raise StopAsyncIteration

            if queue_task in done:
                item = queue_task.result()
                self.task_done()
                if item is not _DEFAULT_VALUE:
                    return item
        except asyncio.CancelledError:
            raise StopAsyncIteration from None
        finally:
            self._closed.clear()

            for task in tasks:
                if not task.done():
                    task.cancel()

    async def get(
        self,
        predicate: typing.Callable[[T], bool] = lambda _: True,
        timeout: float | None = None,
    ) -> T:
        """
        Request an upcoming value that satisfies the `predicate`.

        If used without `timeout` this will block indefinitely until a value satisfies the `predicate`.

        Args:
          predicate: A filter predicate to apply.
          timeout: How many seconds to wait for new value before timing out.

        Raises:
          TimeoutError: If the `timeout` is exceeded.
        """

        start_time = time.perf_counter()

        while True:
            wait_for = timeout + start_time - time.perf_counter() if timeout is not None else None
            try:
                value = await asyncio.wait_for(super().get(), timeout=wait_for)
                self.task_done()
            except (TimeoutError, asyncio.TimeoutError):
                raise TimeoutError from None

            if value is not _DEFAULT_VALUE and predicate(value):
                return value



import dataclasses

from sila import Integer, List, Real, String
from unitelabs.cdk.sila.common.feature import Feature
from unitelabs.cdk.sila.data_types.custom_data_type import CustomDataType


class TestFromNative:
    async def test_should_parse_basic_from_native(self):
        # Create custom data type definition
        @dataclasses.dataclass
        class TestCustomDataType(CustomDataType):
            test_custom_data_type: str

        feature = Feature()
        data_type = TestCustomDataType.attach(feature)

        # Convert from native
        value = await data_type.from_native(feature.context, "Hello, World!")

        # Assert that the method returns the correct value
        assert value == data_type(value=String(value="Hello, World!"))

    async def test_should_parse_basic_from_custom_data_type(self):
        # Create custom data type definition
        @dataclasses.dataclass
        class TestCustomDataType(CustomDataType):
            test_custom_data_type: str

        feature = Feature()
        data_type = TestCustomDataType.attach(feature)
        native = TestCustomDataType(test_custom_data_type="Hello, World!")

        # Convert from native
        value = await data_type.from_native(feature.context, native)

        # Assert that the method returns the correct value
        assert value == data_type(value=String(value="Hello, World!"))

    async def test_should_parse_structure_from_native(self):
        # Create custom data type definition
        @dataclasses.dataclass
        class TestCustomDataType(CustomDataType):
            value: int
            another_value: list[float]

        feature = Feature()
        data_type = TestCustomDataType.attach(feature)

        # Convert from native
        value = await data_type.from_native(feature.context, {"value": 42, "another_value": [1.1, 2.2, 3.3]})

        # Assert that the method returns the correct value
        assert value == data_type(
            data_type.data_type(
                {
                    "value": Integer(value=42),
                    "another_value": List.create(Real)(value=[Real(value=1.1), Real(value=2.2), Real(value=3.3)]),
                }
            )
        )

    async def test_should_parse_structure_from_custom_data_type(self):
        # Create custom data type definition
        @dataclasses.dataclass
        class TestCustomDataType(CustomDataType):
            value: int
            another_value: list[float]

        feature = Feature()
        data_type = TestCustomDataType.attach(feature)
        native = TestCustomDataType(value=42, another_value=[1.1, 2.2, 3.3])

        # Convert from native
        value = await data_type.from_native(feature.context, native)

        # Assert that the method returns the correct value
        assert value == data_type(
            data_type.data_type(
                {
                    "value": Integer(value=42),
                    "another_value": List.create(Real)(value=[Real(value=1.1), Real(value=2.2), Real(value=3.3)]),
                }
            )
        )


class TestToNative:
    async def test_should_parse_basic(self):
        # Create custom data type definition
        @dataclasses.dataclass
        class TestCustomDataType(CustomDataType):
            test_custom_data_type: str

        feature = Feature()
        data_type = TestCustomDataType.attach(feature)
        value = await data_type.from_native(feature.context, "Hello, World!")

        # Convert to native
        native = await value.to_native(feature.context)

        # Assert that the method returns the correct value
        assert isinstance(native, TestCustomDataType)
        assert native.test_custom_data_type == "Hello, World!"

    async def test_should_parse_structure(self):
        # Create custom data type definition
        @dataclasses.dataclass
        class TestCustomDataType(CustomDataType):
            value: int
            another_value: list[float]

        feature = Feature()
        data_type = TestCustomDataType.attach(feature)
        value = await data_type.from_native(feature.context, {"value": 42, "another_value": [1.1, 2.2, 3.3]})

        # Convert to native
        native = await value.to_native(feature.context)

        # Assert that the method returns the correct value
        assert isinstance(native, TestCustomDataType)
        assert native.value == 42
        assert native.another_value == [1.1, 2.2, 3.3]



from .intermediate import Intermediate
from .observable_command import ObservableCommand
from .status import Status
from .unobservable_command import UnobservableCommand

__all__ = ["Intermediate", "ObservableCommand", "Status", "UnobservableCommand"]
